/* A ceelog event.

Copyright (C) 2012 Red Hat, Inc. All rights reserved.
This copyrighted material is made available to anyone wishing to use, modify,
copy, or redistribute it subject to the terms and conditions of the GNU General
Public License v.2.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.

Author: Miloslav Trmač <mitr@redhat.com> */
#include <config.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>

#include <glib.h>
#include <json-c/json_object.h>
#include <json-c/json_tokener.h>

#include <ceelog/event.h>

#include "event_private.h"

 /* Event data. */

/* A ceelog event.
   This will eventually need to be modified to support BSON as well. */
struct ceelog_event
{
  int flags;			/* EVT_* */
  char *text;			/* Original full text, For g_free () */
  size_t text_len;

  /* Parsed contents below: strings point into "text" above.
     If EVT_HAVE_PARSED_TEXT, the fields have provisional values that haven't
     accounted for data in JSON. */
  int pri;			/* If EVT_HAVE_PRI; -1 if not available */
  time_t timestamp;    /* If EVT_HAVE_TIMESTAMP; (time_t)-1 if not available */
  /* If EVT_HAVE_HOSTNAME; NULL if not available.  Contains unmodified hostname
     or IP address from the event, without any canonicalization. */
  const char *hostname_start;
  size_t hostname_len;		/* If EVT_HAVE_HOSTNAME */
  const char *tag_start;	/* If EVT_HAVE_TAG; NULL if not available */
  size_t tag_len;		/* If EVT_HAVE_MSG */
  /* Note that PID, if any, is not stored here (for now?). */
  /* If EVT_HAVE_MSG.  Leading ":" and space, if any, is stripped! */
  const char *msg_start;
  size_t msg_len;		/* If EVT_HAVE_MSG */
  /* If EVT_HAVE_PARSED_TEXT; NULL if not available. For json_object_put (). */
  struct json_object *json;
};

/* Parsing of the event happens on-demand, this indicates whether fields have
   been processed. */
enum {
  EVT_HAVE_PARSED_TEXT = 0x01,
  EVT_HAVE_PRI = 0x02,
  EVT_HAVE_TIMESTAMP = 0x04,
  EVT_HAVE_HOSTNAME = 0x08,
  EVT_HAVE_TAG = 0x10,
  EVT_HAVE_MSG = 0x20,
};

 /* Text parsing, based on RFC 3164 */

/* Parse PRI field, if any, starting at START up to END, into EVENT.
   Return a pointer to the following character. */
static const char *
parse_pri (struct ceelog_event *event, const char *start, const char *end)
{
  int value;
  const char *p;

  event->pri = -1;
  if (start == end || *start != '<')
    return start;

  value = 0;
  p = start + 1;
  for (;;)
    {
      if (p == end || p - start == 4)
	return start;
      if (*p < '0' || *p > '9')
	break;
      value = value * 10 + *p - '0';
      p++;
    }
  assert (p < end);
  if (p == start + 1 || *p != '>')
    return start;

  /* We have a valid number.  Check for prohibited leading zeroes.
     (Maximum of "value" is not checked because glibc's LOG_FACMASK
     suggests there might be more facilities than 0...23.) */
  assert (start + 2 < end);
  if (value != 0)
    {
      if (start[1] == '0')
	return start;
    }
  else
    {
      if (start[2] == '0')
	return start;
    }

  event->pri = value;
  return p + 1;
}

/* Parse a three-character month abbreviation at P.
   Return a month number for tm_mon or -1 on error. */
static int
parse_month_abbr (const char *p)
{
  switch (p[0])
    {
#define CASE(C1, C2, VAL)			\
      if (p[1] == (C1) && p[2] == (C2))		\
	  return (VAL);

    case 'A':
      CASE ('p', 'r', 3);
      CASE ('u', 'g', 7);
      return -1;

    case 'D':
      CASE ('e', 'c', 11);
      return -1;

    case 'F':
      CASE ('e', 'b', 1);
      return -1;

    case 'J':
      CASE ('a', 'n', 0);
      CASE ('u', 'n', 5);
      CASE ('u', 'l', 6);
      return -1;

    case 'M':
      CASE ('a', 'r', 2);
      CASE ('a', 'y', 4);
      return -1;

    case 'N':
      CASE ('o', 'v', 10);
      return -1;

    case 'O':
      CASE ('c', 't', 9);
      return -1;

    case 'S':
      CASE ('e', 'p', 8);
      return -1;

    default:
      return -1;
#undef CASE
    }
}

/* Parse TIMESTAMP field, if any, starting at START up to END, into EVENT.
   Return a pointer to the following character. */
static const char *
parse_timestamp (struct ceelog_event *event, const char *start, const char *end)
{
  time_t t;
  struct tm current_tm, tm;
  long l;
  char *num_end;
  const char *timestamp_end;

  event->timestamp = -1;
  /* Mmm dd hh:mm:ss
     0123456789012345 */
  if (end - start < 15 || start[3] != ' ' || start[6] != ' ' || start[9] != ':'
      || start[12] != ':' || start[15] != ' ')
    return start;

  tm.tm_mon = parse_month_abbr (start);
  if (tm.tm_mon == -1)
    return start;

  /* We do not enforce the requirement to use use leading spaces/zeros */
  errno = 0;
  /* Terminates at start + 6 at most, see above */
  l = strtol (start + 4, &num_end, 10);
  if (errno != 0 || num_end != start + 6 || l < 1 || l > 31)
    return start;
  tm.tm_mday = l;

  errno = 0;
  /* Terminates at start + 9 at most, see above */
  l = strtol (start + 7, &num_end, 10);
  if (errno != 0 || num_end != start + 9 || l < 0 || l > 23)
    return start;
  tm.tm_hour = l;

  errno = 0;
  /* Terminates at start + 12 at most, see above */
  l = strtol (start + 10, &num_end, 10);
  if (errno != 0 || num_end != start + 12 || l < 0 || l > 59)
    return start;
  tm.tm_min = l;

  /* Ensure that strtol () below will terminate at start[15] at most, because
     the numeric sequence starts earlier, and we know start[15] == ' '. */
  if (start[14] < '0' || start[14] > '9')
    return start;
  errno = 0;
  l = strtol (start + 13, &num_end, 10);
  /* Note that allowing values 60 and 61 contradicts RFC 3164. */
  if (errno != 0 || num_end != start + 15 || l < 0 || l > 61)
    return start;
  tm.tm_sec = l;

  /* Computing the timestamp value can theoretically fail; in that case, don't
     "return start" as in other failure cases - the timestamp text is
     actually fine, it's the OS that is in trouble. */
  timestamp_end = start + 16;

  /* All this only to get the year information.

     FIXME? cache this somehow.  For example, we could compute the start of
     the next year and avoid the localtime() calls, at least. */
  t = time (NULL);
  if (t == -1)
    return timestamp_end;
  tzset (); /* FIXME: is this necessary? */
  if (localtime_r (&t, &current_tm) == NULL)
    return timestamp_end;
  tm.tm_year = current_tm.tm_year;
  if (tm.tm_mon > current_tm.tm_mon
      || (tm.tm_mon == current_tm.tm_mon
	  && (tm.tm_mday > current_tm.tm_mday
	      || (tm.tm_mday == current_tm.tm_mday
		  && (tm.tm_hour > current_tm.tm_hour
		      || (tm.tm_hour == current_tm.tm_hour
			  && (tm.tm_min > current_tm.tm_min
			      || (tm.tm_min == current_tm.tm_min
				  && tm.tm_sec > current_tm.tm_sec))))))))
    tm.tm_year--;
  tm.tm_isdst = -1; /* Unfortunate */

  event->timestamp = mktime (&tm); /* May be -1 on error */

  return timestamp_end;
}

/* Parse HOSTNAME field, if any, starting at START up to END, into EVENT.
   Return a pointer to the following character. */
static const char *
parse_hostname (struct ceelog_event *event, const char *start, const char *end)
{
  const char *p;

  event->hostname_start = NULL;

  p = start;
  for (;;)
    {
      if (p == end)
	return start;
      if (*p == ' ')
	break;
      if (!isascii((unsigned char)*p))
	return start;
      p++;
    }

  if (p == start)
    return start;

  event->hostname_start = start;
  event->hostname_len = p - start;
  return p + 1;
}

/* Parse TAG field, if any, starting at START up to END, into EVENT.
   Return a pointer to the following character. */
static const char *
parse_tag (struct ceelog_event *event, const char *start, const char *end)
{
  const char *p;

  event->tag_start = NULL;

  for (p = start; p < end && g_ascii_isalnum(*p); p++)
    ;
  /* Prefer a non-empty MSG to a non-empty TAG if there is only one of them */
  if (p == start || p == end)
    return start;

  event->tag_start = start;
  event->tag_len = p - start;
  return p;
}

/* Parse PID field, if any, related to EVENT, starting at START up to END.
   Return a pointer to the following character. */
static const char *
skip_pid (struct ceelog_event *event, const char *start, const char *end)
{
  const char *p;

  /* The PID event is not exactly specified by RFC 3164; assume it can't be
     present without a tag. */
  if (event->tag_start == NULL || start == end || *start != '[')
    return start;

  p = start + 1;
  for (;;)
    {
      if (p == end)
	return start;
      if (*p < '0' || *p > '9')
	break;
      p++;
    }
  assert (p < end);
  if (p == start + 1 || *p != ']')
    return start;

  return p + 1;
}

/* Return a JSON object corresponding to STRING with LEN, or NULL on error */
static struct json_object *
parse_json (const char *string, size_t len)
{
  struct json_tokener *tok;
  struct json_object *obj;

  /* FIXME: can we reuse a single tokener object, or perhaps a per-thread
     one? */
  tok = json_tokener_new ();
  if (tok == NULL)
    return NULL;

  obj = json_tokener_parse_ex (tok, string, len);
  if (obj != NULL
      && (tok->err != json_tokener_success || (size_t)tok->char_offset < len
	  || !json_object_is_type (obj, json_type_object)))
    {
      /* Should we warn about it, at least in some kind of verbose mode? */
      json_object_put (obj);
      obj = NULL;
    }

  json_tokener_free (tok);

  return obj;
}

/* Parse EVENT->text, and set EVT_HAVE_PARSED_TEXT */
static void
ceelog_event_parse_text (struct ceelog_event *event)
{
  static const char cee_marker[5] = "@cee:";

  const char *p, *end;

  if ((event->flags & EVT_HAVE_PARSED_TEXT) != 0)
    return;

  p = event->text;
  end = event->text + event->text_len;

  p = parse_pri (event, p, end);
  p = parse_timestamp (event, p, end);
  p = parse_hostname (event, p, end);
  p = parse_tag (event, p, end);
  p = skip_pid (event, p, end);

  if (p < end && *p == ':')
    p++;
  if (p < end && *p == ' ')
    p++;
  event->msg_start = p;
  event->msg_len = end - p;

  if ((size_t)(end - p) > sizeof (cee_marker)
      && memcmp (p, cee_marker, sizeof (cee_marker)) == 0)
    {
      const char *json_start;

      json_start = p + sizeof (cee_marker);
      event->json = parse_json (json_start, end - json_start);
    }
  else
    event->json = NULL;

  event->flags |= EVT_HAVE_PARSED_TEXT;
}

 /* Cached property access */

/* Allocate and return a new event from STRING of LEN bytes.
   FIXME: should this be a public API, or do we want to restrict ouselves to
   provided sources? */
struct ceelog_event *
ceelog_event_new_from_string (const char *string, size_t len)
{
  struct ceelog_event *e;

  e = g_slice_new (struct ceelog_event);
  e->flags = 0;
  e->text = g_strndup (string, len);
  e->text_len = len;
  return e;
}

/* Free EVENT. */
void
ceelog_event_free (struct ceelog_event *event)
{
  g_return_if_fail (event != NULL);

  g_free (event->text);
  if ((event->flags & EVT_HAVE_PARSED_TEXT) != 0 && event->json != NULL)
    json_object_put (event->json);
  g_slice_free (struct ceelog_event, event);
}

/* Return a syslog facility of EVENT, or -1 if not available. */
int
ceelog_event_get_facility (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, -1);

  if ((event->flags & EVT_HAVE_PRI) == 0)
    {
      ceelog_event_parse_text (event);
      /* FIXME: Parse JSON data as an alternative */
      event->flags |= EVT_HAVE_PRI;
    }

  assert ((event->flags & EVT_HAVE_PRI) != 0);
  if (event->pri != -1)
    return event->pri & LOG_FACMASK;
  return -1;
}

/* Return a syslog priority of EVENT, or -1 if not available. */
int
ceelog_event_get_priority (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, -1);

  if ((event->flags & EVT_HAVE_PRI) == 0)
    {
      ceelog_event_parse_text (event);
      /* FIXME: Parse JSON data as an alternative */
      event->flags |= EVT_HAVE_PRI;
    }

  assert ((event->flags & EVT_HAVE_PRI) != 0);
  if (event->pri != -1)
    return event->pri & LOG_PRIMASK;
  return -1;
}

/* Return a timestamp EVENT, or -1 if not available. */
time_t
ceelog_event_get_timestamp (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, -1);

  if ((event->flags & EVT_HAVE_TIMESTAMP) == 0)
    {
      ceelog_event_parse_text (event);
      /* FIXME: Parse JSON data as an alternative */
      event->flags |= EVT_HAVE_TIMESTAMP;
    }

  assert ((event->flags & EVT_HAVE_TIMESTAMP) != 0);
  return event->timestamp;
}

/* Return host name of EVENT (for g_free ()), or NULL if not available.
   This is the unmodified host name or IP address from the event, without any
   canonicalization. */
char *
ceelog_event_get_hostname (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, NULL);

  if ((event->flags & EVT_HAVE_HOSTNAME) == 0)
    {
      ceelog_event_parse_text (event);
      /* FIXME: Parse JSON data as an alternative */
      event->flags |= EVT_HAVE_HOSTNAME;
    }

  assert ((event->flags & EVT_HAVE_HOSTNAME) != 0);
  if (event->hostname_start != NULL)
    return g_strndup (event->hostname_start, event->hostname_len);
  return NULL;
}

/* Return tag of EVENT (for g_free ()), or NULL if not available.
   PID, if any, is not included. */
char *
ceelog_event_get_tag (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, NULL);

  if ((event->flags & EVT_HAVE_TAG) == 0)
    {
      ceelog_event_parse_text (event);
      /* FIXME: Parse JSON data as an alternative */
      event->flags |= EVT_HAVE_TAG;
    }

  assert ((event->flags & EVT_HAVE_TAG) != 0);
  if (event->tag_start != NULL)
    return g_strndup (event->tag_start, event->tag_len);
  return NULL;
}

/* Ensure EVT_HAVE_MSG is set on EVENT */
static void
ensure_message (struct ceelog_event *event)
{
  if ((event->flags & EVT_HAVE_MSG) == 0)
    {
      ceelog_event_parse_text (event);
      if (event->json != NULL)
	{
	  struct json_object *msg;

	  msg = json_object_object_get (event->json, "msg");
	  if (msg != NULL && json_object_is_type(msg, json_type_string))
	    {
	      event->msg_start = json_object_get_string(msg);
	      /* Future JSON versions have json_object_get_string_len() */
	      event->msg_len = strlen(event->msg_start);
	    }
	}
      event->flags |= EVT_HAVE_MSG;
    }
}

/* Return a pointer to the message of EVENT, and set LEN to its length.
   For JSON messages with a "msg" field, returns the "msg" field only.
   For any other messages, this contains all text after the tag and PID, if any,
   with the leading ':' and ' ' stripped. */
const char *
ceelog_event_get_message_pointer (struct ceelog_event *event, size_t *len)
{
  ensure_message (event);
  assert ((event->flags & EVT_HAVE_MSG) != 0);
  if (event->msg_start != NULL)
    *len = event->msg_len;
  return event->msg_start;
}

/* Return the message of EVENT (for g_free ()).
   For JSON messages with a "msg" field, returns the "msg" field only.
   For any other messages, this contains all text after the tag and PID, if any,
   with the leading ':' and ' ' stripped. */
char *
ceelog_event_get_message (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, NULL);

  ensure_message (event);
  assert ((event->flags & EVT_HAVE_MSG) != 0);
  return g_strndup (event->msg_start, event->msg_len);
}

/* Return the JSON data in EVENT (for json_object_put ()), or NULL if not
   present.
   The JSON data is unmodified (metadata in the text part is not added,
   JSON variants of the text part metadata is not removed).  The caller must
   not modify the data. */
struct json_object *
ceelog_event_get_json (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, NULL);

  ceelog_event_parse_text (event);
  if (event->json != NULL)
    return json_object_get (event->json);
  return NULL;
}

/* Return a best possible representation of EVENT as a string, for g_free ().
   The string does not end with any line terminator.  Preserving the original
   representation is preferred to converting the data into an ideal canonical
   format. */
char *
ceelog_event_to_string (struct ceelog_event *event)
{
  g_return_val_if_fail (event != NULL, NULL);

  return g_memdup (event->text, event->text_len + 1);
}
