/* A ceelog expression creation and parsing.

Copyright (C) 2012 Red Hat, Inc. All rights reserved.
This copyrighted material is made available to anyone wishing to use, modify,
copy, or redistribute it subject to the terms and conditions of the GNU General
Public License v.2.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.

Author: Miloslav Trmač <mitr@redhat.com> */
#include <config.h>

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <json-c/json_object.h>
#include <pcre.h>

#include <ceelog/ceelog.h>
#include <ceelog/expression.h>

#include "expression_private.h"

 /* Regex utils */

/* Compile REGEX into DEST, or set ERROR.
   Return 0 if OK, -1 on error */
int
ceelog_regex_compile (struct ceelog_regex *dest, const char *regex,
		      GError **error)
{
  int offset;
  const char *msg;
  pcre *re;
  pcre_extra *extra;

  /* FIXME? support flags:
     g ==
     i == PCRE_CASELESS
     m == PCRE_MULTILINE, case-insensitive, extended */
  re = pcre_compile (regex,
		     PCRE_DOLLAR_ENDONLY | PCRE_JAVASCRIPT_COMPAT | PCRE_UTF8,
		     &msg, &offset, NULL);
  if (re == NULL)
    {
      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Invalid regex `%s': %s"), regex, msg);
      return -1;
    }

  extra = pcre_study (re, 0, &msg);
  if (msg != NULL)
    {
      pcre_free (re);
      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Error compiling regex `%s': %s"), regex, msg);
      return -1;
    }

  dest->re = re;
  dest->extra = extra;
  return 0;
}

/* Free RE */
G_GNUC_INTERNAL
void
ceelog_regex_free (struct ceelog_regex *re)
{
  pcre_free_study (re->extra);
  pcre_free (re->re);
}

 /* Expression creation */

/* Free EXPR, including all its subexpressions. */
void
ceelog_expr_free (struct ceelog_expr *expr)
{
  g_return_if_fail (expr != NULL);

  switch (expr->op)
    {
    case IO_FIELD_EQ: case IO_FIELD_NE: case IO_FIELD_MATCHES:
    case IO_FIELD_DOESNT_MATCH:
      g_free (expr->v.cmp.field_path);
      /* Fall through */
    case IO_UNSTRUCTURED_MATCHES: case IO_UNSTRUCTURED_DOESNT_MATCH:
      g_free (expr->v.cmp.value);
      break;
    case IO_AND: case IO_OR:
      ceelog_expr_free (expr->v.sub[1]);
      /* Fall through */
    case IO_NOT:
      ceelog_expr_free (expr->v.sub[0]);
      break;

    default:
      g_return_if_reached ();
    }
  g_free (expr);
}

/* Create a comparison expression that compares FIELD_PATH using OP with VALUE.
   Return the created expression, or NULL on error.

   Allowed OP values are CEELOG_EO_{EQ,NE,MATCHES,DOESNT_MATCH}.
   If an expression doesn't contain FIELD_PATH, the expression is false
   regardless of the value of OP. */
struct ceelog_expr *
ceelog_expr_new_comparison (const char *field_path, enum ceelog_expr_op op,
			    const char *value, GError **error)
{
  struct ceelog_expr *expr;

  g_return_val_if_fail (field_path != NULL, NULL);
  /* op is validated later */
  g_return_val_if_fail (value != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  expr = g_new (struct ceelog_expr, 1);
  switch (op)
    {
    case CEELOG_EO_EQ:
      expr->op = IO_FIELD_EQ;
      goto cmp_value;

    case CEELOG_EO_NE:
      expr->op = IO_FIELD_NE;
    cmp_value:
      expr->v.cmp.field_path = g_strdup (field_path);
      expr->v.cmp.value = g_strdup (value);
      break;

    case CEELOG_EO_MATCHES:
      expr->op = IO_FIELD_MATCHES;
      goto cmp_regex;

    case CEELOG_EO_DOESNT_MATCH:
      expr->op = IO_FIELD_DOESNT_MATCH;
    cmp_regex:
      {
	struct ceelog_regex re;

	/* Check syntax now for better error reporting; the regex will be
	   recompiled in ceelog_compile_expr (). */
	if (ceelog_regex_compile (&re, value, error) != 0)
	  {
	    g_free (expr);
	    return NULL;
	  }
	ceelog_regex_free (&re);
	expr->v.cmp.field_path = g_strdup (field_path);
	expr->v.cmp.value = g_strdup (value);
	break;
      }

    default:
      g_free (expr);
      g_return_val_if_reached (NULL);
    }

  return expr;
}

/* Create a regex expression that matches a non-JSON event against REGEX
   using OP.
   Return the created expression, or NULL on error.

   Allowed OP values are CEELOG_EO_{MATCHES,DOESNT_MATCH}.
   Note that this is intended to be used only for unstructured events, but they
   may also be expressed in a structured representation; in that case, the
   regex is used to match the "msg" field. */
struct ceelog_expr *
ceelog_expr_new_unstructured_regex (enum ceelog_expr_op op, const char *regex,
				    GError **error)
{
  struct ceelog_regex re;
  struct ceelog_expr *expr;

  /* op is validated later */
  g_return_val_if_fail (regex != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  /* Check syntax now for better error reporting; the regex will be
     recompiled in ceelog_compile_expr (). */
  if (ceelog_regex_compile (&re, regex, error) != 0)
    return NULL;
  ceelog_regex_free (&re);

  expr = g_new (struct ceelog_expr, 1);
  switch (op)
    {
    case CEELOG_EO_MATCHES:
      expr->op = IO_UNSTRUCTURED_MATCHES;
      break;

    case CEELOG_EO_DOESNT_MATCH:
      expr->op = IO_UNSTRUCTURED_DOESNT_MATCH;
      break;

    default:
      g_free (expr);
      g_return_val_if_reached (NULL);
    }

  expr->v.cmp.value = g_strdup (regex);
  return expr;
}

/* Create a binary expression for OP and subexpressions E1 and E2.
   Return the created expression.
   Allowed OP values are CEELOG_EO_{AND,OR}.
   Takes ownership of E1 and E2, the caller must not free them. */
struct ceelog_expr *
ceelog_expr_new_binary (enum ceelog_expr_op op, struct ceelog_expr *e1,
			struct ceelog_expr *e2)
{
  struct ceelog_expr *expr;

  /* op is validated later */
  g_return_val_if_fail (e1 != NULL, NULL);
  g_return_val_if_fail (e2 != NULL, NULL);

  expr = g_new (struct ceelog_expr, 1);
  switch (op)
    {
    case CEELOG_EO_AND:
      expr->op = IO_AND;
      break;

    case CEELOG_EO_OR:
      expr->op = IO_OR;
      break;

    default:
      g_free (expr);
      g_return_val_if_reached (NULL);
    }

  expr->v.sub[0] = e1;
  expr->v.sub[1] = e2;
  return expr;
}

/* Create an unary expression for OP and subexpression SUB.
   Return the created expression.
   The only allowed OP value is CEELOG_EO_NOT.
   Takes ownership of SUB, the caller must not free it. */
struct ceelog_expr *
ceelog_expr_new_unary (enum ceelog_expr_op op, struct ceelog_expr *sub)
{
  struct ceelog_expr *expr;

  g_return_val_if_fail (op == CEELOG_EO_NOT, NULL);
  g_return_val_if_fail (sub != NULL, NULL);

  expr = g_new (struct ceelog_expr, 1);
  expr->op = IO_NOT;
  expr->v.sub[0] = sub;
  return expr;
}

 /* Expression parsing */

/* The formal grammar:

   start: or-expression

   or-expression: and-expression
   or-expression: or-expression || and-expression

   and-expression: primary-expression
   and-expression: and-expression && primary-expression

   primary-expression: ! primary-expression
   primary-expression: ( or-expression )
   primary-expression: field-expression
   primary-expression: unstructured-expression

   field-expression: field-path string-op string
   field-expression: field-path regex-op regex
   unstructured-expression: regex

   field-path: identifier
   field-path: identifier ! field-path */

enum token_type
  {
    T_EQ, T_NE, T_MATCHES, T_DOESNT_MATCH, T_AND, T_OR, T_NOT, T_LEFT_PAREN,
    T_RIGHT_PAREN, T_STRING, T_REGEX, T_IDENTIFIER,
    T_UNKNOWN, T_EOF
  };

/* Expression parsing status */
struct parsing
{
  GError **error;
  enum token_type token;
  const char *token_start;     /* Start of token in expression source */
  int token_len;	       /* int because it must be usable in %.*s */
  char *token_value;	       /* T_STRING, T_REGEX, T_IDENTIFIER;
				  for g_free () */
  const char *src;	       /* Expression source, after the current token */
};

/* Discard P->token_value, if any, and parse the next token in P->src.
   On success, return 0.
   On error, set P->error and return -1. */
static int
lex (struct parsing *p)
{
  const char *src;

  g_free (p->token_value);
  p->token_value = NULL;

  src = p->src;
 again:
  while (*src == ' ' || *src == '\t' || *src == '\n' || *src == '\r'
	 || *src == '\v' || *src == '\f')
    src++;

  p->token_start = src;
  switch (*src)
    {
    case '\0':
      p->token = T_EOF;
      break;

    case '!':
      src++;
      if (*src == '=')
	{
	  src++;
	  p->token = T_NE;
	  break;
	}
      if (*src == '~')
	{
	  src++;
	  p->token = T_DOESNT_MATCH;
	  break;
	}
      p->token = T_NOT;
      break;

    case '/':
      if (src[1] == '/')
	{
	  for (src += 2; *src != '\0' && *src != '\n' && *src != '\r';
	       src++)
	    ;
	  goto again;
	}
      if (src[1] == '*')
	{
	  src += 2;
	  do
	    {
	      while (*src != '*')
		{
		  if (*src == '\0')
		    {
		      g_set_error (p->error, CEELOG_ERROR,
				   CEELOG_ERROR_EXPRESSION_SYNTAX,
				   _("Unterminated comment"));
		      return -1;
		    }
		  src++;
		}
	      src++;
	    }
	  while (*src != '/');
	  src++;
	  goto again;
	}
      /* Fall through */
    case '"':
      {
	GString *buf;
	char delimiter;

	delimiter = *src;
	src++;

	buf = g_string_new ("");
	while (*src != delimiter)
	  {
	    if (*src == '\0')
	      {
		g_set_error (p->error, CEELOG_ERROR,
			     CEELOG_ERROR_EXPRESSION_SYNTAX,
			     _("Terminating delimiter missing"));
		g_string_free (buf, TRUE);
		return -1;
	      }
	    if (*src == '\\')
	      {
		if (src[1] == '\0')
		  {
		    g_set_error (p->error, CEELOG_ERROR,
				 CEELOG_ERROR_EXPRESSION_SYNTAX,
				 _("Terminating delimiter missing"));
		    g_string_free (buf, TRUE);
		    return -1;
		  }
		if (delimiter == '"')
		  {
		    src++;
		    // FIXME? Implement standard JavaScript escape sequences for
		    // strings?
		    if (*src != '\\' && *src != delimiter)
		      {
			g_set_error (p->error, CEELOG_ERROR,
				     CEELOG_ERROR_EXPRESSION_SYNTAX,
				     "Unknown escape sequence `\\%c'", *src);
			g_string_free (buf, TRUE);
			return -1;
		      }
		  }
		else
		  {
		    /* Store the escape sequence without modification */
		    buf = g_string_append_c (buf, *src);
		    src++;
		  }
	      }
	    buf = g_string_append_c (buf, *src);
	    src++;
	  }
	src++;

	p->token_value = g_string_free (buf, FALSE);
	p->token = delimiter == '/' ? T_REGEX : T_STRING;
	break;
      }

    case '&':
      src++;
      if (*src == '&')
	{
	  src++;
	  p->token = T_AND;
	  break;
	}
      p->token = T_UNKNOWN;
      break;

    case '(':
      src++;
      p->token = T_LEFT_PAREN;
      break;

    case ')':
      src++;
      p->token = T_RIGHT_PAREN;
      break;

    case '=':
      src++;
      if (*src == '=')
	{
	  src++;
	  p->token = T_EQ;
	  break;
	}
      p->token = T_UNKNOWN;
      break;

    case '|':
      src++;
      if (*src == '|')
	{
	  src++;
	  p->token = T_OR;
	  break;
	}
      p->token = T_UNKNOWN;
      break;

    case '~':
      src++;
      p->token = T_MATCHES;
      break;

    default:
      if (g_ascii_isalpha (*src) || *src == '_')
	{
	  do
	    src++;
	  while (g_ascii_isalnum (*src) || *src == '_');
	  p->token_value = g_strndup (p->token_start, src - p->token_start);
	  p->token = T_IDENTIFIER;
	  break;
	}
      src++;
      p->token = T_UNKNOWN;
      break;
    }

  if (src - p->token_start > INT_MAX)
    {
      g_set_error (p->error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Token too long"));
      return -1;
    }
  p->src = src;
  p->token_len = src - p->token_start;
  return 0;
}

static struct ceelog_expr *parse_or (struct parsing *p);

/* Parse a field-expression string in *P.
   On success, return the parsed primary-expression.
   On error, set P->error and return NULL. */
static struct ceelog_expr *
parse_field_expr (struct parsing *p)
{
  GString *field_path;
  struct ceelog_expr *e;
  enum ceelog_expr_op op;
  enum token_type value_token;

  g_assert (p->token == T_IDENTIFIER);

  field_path = g_string_new (p->token_value);
  if (lex (p) != 0)
    goto err_field_path;

  while (p->token == T_NOT)
    {
      field_path = g_string_append_c (field_path, '!');
      if (lex (p) != 0)
	goto err_field_path;

      if (p->token != T_IDENTIFIER)
	{
	  g_set_error (p->error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		       _("Expected identifier, got `%.*s'"), p->token_len,
		       p->token_start);
	  goto err_field_path;
	}
      field_path = g_string_append (field_path, p->token_value);
      if (lex (p) != 0)
	goto err_field_path;
    }

  switch (p->token)
    {
    case T_EQ:
      op = CEELOG_EO_EQ;
      value_token = T_STRING;
      break;

    case T_NE:
      op = CEELOG_EO_NE;
      value_token = T_STRING;
      break;

    case T_MATCHES:
      op = CEELOG_EO_MATCHES;
      value_token = T_REGEX;
      break;

    case T_DOESNT_MATCH:
      op = CEELOG_EO_DOESNT_MATCH;
      value_token = T_REGEX;
      break;

    default:
      g_set_error (p->error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Expected operator, got `%.*s'"), p->token_len,
		   p->token_start);
      goto err_field_path;
    }
  if (lex (p) != 0)
    goto err_field_path;

  if (p->token != value_token)
    {
      const char *fmt;

      switch (value_token)
	{
	case T_STRING:
	  fmt = _("Expected string, got `%.*s'");
	  break;

	case T_REGEX:
	  fmt = _("Expected regex, got `%.*s'");
	  break;

	default:
	  g_assert_not_reached ();
	}
      g_set_error (p->error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX, fmt,
		   p->token_len, p->token_start);
      goto err_field_path;
    }
  e = ceelog_expr_new_comparison (field_path->str, op, p->token_value,
				  p->error);
  if (e == NULL)
    goto err_field_path;
  if (lex (p) != 0)
    goto err_e;

  g_string_free (field_path, TRUE);
  return e;

 err_e:
  ceelog_expr_free (e);
 err_field_path:
  g_string_free (field_path, TRUE);
  return NULL;
}

/* Parse an unstructured-expression string in *P.
   On success, return the parsed primary-expression.
   On error, set P->error and return NULL. */
static struct ceelog_expr *
parse_unstructured (struct parsing *p)
{
  struct ceelog_expr *e;

  g_assert (p->token == T_REGEX);

  e = ceelog_expr_new_unstructured_regex (CEELOG_EO_MATCHES, p->token_value,
					  p->error);
  if (e == NULL)
    return NULL;

  if (lex (p) != 0)
    goto err_e;

  return e;

 err_e:
  ceelog_expr_free (e);
  return NULL;
}

/* Parse a primary-expression string in *P.
   On success, return the parsed primary-expression.
   On error, set P->error and return NULL. */
static struct ceelog_expr *
parse_primary (struct parsing *p)
{
  struct ceelog_expr *e;

  switch (p->token)
    {
    case T_NOT:
      if (lex (p) != 0)
	return NULL;
      e = parse_primary (p);
      if (e == NULL)
	return NULL;
      return ceelog_expr_new_unary (CEELOG_EO_NOT, e);

    case T_LEFT_PAREN:
      if (lex (p) != 0)
	return NULL;
      e = parse_or (p);
      if (e == NULL)
	return NULL;
      if (p->token != T_RIGHT_PAREN)
	{
	  g_set_error (p->error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		       _("Right paren expected, got `%.*s'"), p->token_len,
		       p->token_start);
	  goto err_e;
	}
      if (lex (p) != 0)
	goto err_e;
      return e;

    case T_IDENTIFIER:
      return parse_field_expr (p);

    case T_REGEX:
      return parse_unstructured (p);

    default:
      g_set_error (p->error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Unexpected token `%.*s'"), p->token_len, p->token_start);
      return NULL;
    }

 err_e:
  ceelog_expr_free (e);
  return NULL;
}

/* Parse an and-expression string in *P.
   On success, return the parsed and-expression.
   On error, set P->error and return NULL. */
static struct ceelog_expr *
parse_and (struct parsing *p)
{
  struct ceelog_expr *res;

  res = parse_primary (p);
  if (res == NULL)
    return NULL;
  while (p->token == T_AND)
    {
      struct ceelog_expr *e2;

      if (lex (p) != 0)
	goto err_res;
      e2 = parse_primary(p);
      if (e2 == NULL)
	goto err_res;
      res = ceelog_expr_new_binary (CEELOG_EO_AND, res, e2);
    }
  return res;

 err_res:
  ceelog_expr_free (res);
  return NULL;
}

/* Parse an or-expression string in *P.
   On success, return the parsed or-expression.
   On error, set P->error and return NULL. */
static struct ceelog_expr *
parse_or (struct parsing *p)
{
  struct ceelog_expr *res;

  res = parse_and (p);
  if (res == NULL)
    return NULL;
  while (p->token == T_OR)
    {
      struct ceelog_expr *e2;

      if (lex (p) != 0)
	goto err_res;
      e2 = parse_and (p);
      if (e2 == NULL)
	goto err_res;
      res = ceelog_expr_new_binary (CEELOG_EO_OR, res, e2);
    }
  return res;

 err_res:
  ceelog_expr_free (res);
  return NULL;
}

/* Create an expression based on a text EXPR.
   Return the created expression, or NULL on error. */
struct ceelog_expr *
ceelog_expr_parse (const char *expr, GError **error)
{
  struct parsing p;
  struct ceelog_expr *res;

  g_return_val_if_fail (expr != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  p.error = error;
  p.token_value = NULL;
  p.src = expr;

  if (lex (&p) != 0)
    goto err;
  if (p.token == T_EOF)
    {
      g_set_error (p.error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Empty expression"));
      goto err;
    }
  res = parse_or (&p);
  if (res != NULL && p.token != T_EOF)
    {
      ceelog_expr_free (res);
      g_set_error (p.error, CEELOG_ERROR, CEELOG_ERROR_EXPRESSION_SYNTAX,
		   _("Unexpected trailing token `%.*s'"), p.token_len,
		   p.token_start);
      goto err;
    }

  g_free (p.token_value);
  return res;

 err:
  g_free (p.token_value);
  return NULL;
}

