/* ceelog_source_new_log_file () implementation.

Copyright (C) 2012 Red Hat, Inc. All rights reserved.
This copyrighted material is made available to anyone wishing to use, modify,
copy, or redistribute it subject to the terms and conditions of the GNU General
Public License v.2.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.

Authors: Miloslav Trmač <mitr@redhat.com>
         Milan Bartos <mbartos@redhat.com> */
#define _GNU_SOURCE

#include <config.h>

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/inotify.h>

#include <glib.h>
#include <glib/gi18n-lib.h>

#include <ceelog/ceelog.h>
#include <ceelog/source.h>

#include "event_private.h"
#include "source_private.h"

/* For now, just use a FILE * and getline().  We may optimize this later if
   necessary. */

/* Close SOURCE and free internal data. */
static void
file_close (struct ceelog_source *source)
{
  g_free (source->v.logfile.prevline);
  inotify_rm_watch (source->v.logfile.inotify_desc,
		    source->v.logfile.watch_desc);
  close (source->v.logfile.inotify_desc); /* close (-1) is harmless. */
  g_free (source->v.logfile.path);
  if (source->v.logfile.file)
    {
      fclose (source->v.logfile.file);
    }
}

/* Read the next event from SOURCE.

   Return the event (for ceelog_event_free ()) if it is available, or blocks
   until next event is available. Or return NULL on error.
  */
#define BUFF_SIZE ((sizeof (struct inotify_event) + NAME_MAX + 1) * 10)
static struct ceelog_event *
file_read_event (struct ceelog_source *source, GError ** error)
{
  ssize_t readlen;
  ssize_t len;
  size_t length = 0;
  unsigned int i;

  char buff[BUFF_SIZE];
  char *helpline = NULL;
  char *tmpline = NULL;
  char *end = NULL;

  struct inotify_event pevent;
  struct ceelog_event *ret = NULL;

  while (1)
    {
      readlen = getline (&helpline, &length, source->v.logfile.file);

      if (readlen == -1)
	{
	  if (ferror (source->v.logfile.file))
	    {
	      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_IO,
			   _("Error reading `%s': %s"), source->v.logfile.path,
			   strerror (errno));
	      goto err_helpline;
	    }
	  /* else no data is available yet, wait for an inotify event. */
	}
      else
	{
	  /* We have some data.  Append it to previous line fragment, if any. */
	  if (source->v.logfile.prevline != NULL)
	    {
	      if (asprintf
		  (&tmpline, "%s%s",
		   source->v.logfile.prevline, helpline) == -1)
		{
		  g_set_error (error, CEELOG_ERROR,
			       CEELOG_ERROR_IO,
			       _("Error appending message fragment"));
		  goto err_helpline;
		}
	      g_free (source->v.logfile.prevline);
	      source->v.logfile.prevline = NULL;
	      free (helpline);
	      helpline = tmpline;
	      readlen = strlen (helpline);
	    }

	  if (memchr (helpline, '\n', readlen) != NULL)
	    {
	      /* We have a complete line.  First NUL byte to make sure embedded
		 NULs don't confuse ceelog_event_new_from_string (). */
	      end = memchr (helpline, '\0', readlen + 1);
	      assert (end != NULL);

	      /* remove newline character from message */
	      if (end > helpline && end[-1] == '\n')
		end--;
	      if (end > helpline && end[-1] == '\r')
		end--;

	      ret = ceelog_event_new_from_string (helpline, end - helpline);
	      free (helpline);
	      return ret;
	    }
	  source->v.logfile.prevline = g_strdup (helpline);
	  /* That's all for now, wait for an inotify event. */
	}

      len = read (source->v.logfile.inotify_desc, buff, BUFF_SIZE);
      if (len < 0)
	{
	  g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_IO,
		       _("Error reading inotify event: %s"),
		       strerror (errno));
	  goto err_helpline;
	}

      assert (len >= 0);
      for (i = 0; i + sizeof (struct inotify_event) <= (size_t)len;
	   i += sizeof (struct inotify_event) + pevent.len)
	{
	  memcpy (&pevent, &buff[i], sizeof (struct inotify_event));
	  if (i + sizeof (struct inotify_event) + pevent.len > (size_t)len)
	    break;

	  if (pevent.mask & IN_MOVE_SELF)
	    {			/* file has been moved (~rotated), reopen the new one */
	      fclose (source->v.logfile.file);
	      sleep (1);
	      source->v.logfile.file = fopen (source->v.logfile.path, "r");

	      if (source->v.logfile.file == NULL)
		{
		  g_set_error (error, CEELOG_ERROR,
			       CEELOG_ERROR_IO,
			       _("Error opening `%s': %s"),
			       source->v.logfile.path, strerror (errno));
		  goto err_helpline;
		}

	      inotify_rm_watch (source->v.logfile.inotify_desc,
				source->v.logfile.watch_desc);
	      source->v.logfile.watch_desc =
		inotify_add_watch (source->v.logfile.inotify_desc,
				   source->v.logfile.path, IN_ALL_EVENTS);
	      if (source->v.logfile.watch_desc < 0)
		{
		  g_set_error (error, CEELOG_ERROR,
			       CEELOG_ERROR_FAILED,
			       _
			       ("Failed to add Inotify watch on `%s': %s"),
			       source->v.logfile.path, strerror (errno));
		  goto err_helpline;
		}
	    }
	}
    }

err_helpline:
  free (helpline);
  return NULL;

}

/* Read the next event from SOURCE.  If no event is currently available,
   fail with CEELOG_ERROR_WOULD_BLOCK.

   Return the event (for ceelog_event_free ()) if it is available, or NULL
   with appropriate error set on error.

   The last line, even if not terminated with newline character, is also returned.
 */
static struct ceelog_event *
file_read_event_nonblocking (struct ceelog_source *source, GError ** error)
{
  ssize_t readlen = 0;
  size_t length = 0;
  char *end = NULL;
  char *readline = NULL;
  struct ceelog_event *ret = NULL;

  readlen = getline (&readline, &length, source->v.logfile.file);

  if (readlen == -1)
    {
      if (feof (source->v.logfile.file))
	{
	  g_set_error (error, CEELOG_ERROR,
		       CEELOG_ERROR_WOULD_BLOCK,
		       _("Reading `%s' would block"), source->v.logfile.path);
	}
      else
	{
	  g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_IO,
		       _("Error reading file `%s'"), source->v.logfile.path);
	}
      free (readline);
      return NULL;
    }

  /* find first nullbyte to avoid ceelog_event_new_from_string confusion */
  end = memchr (readline, '\0', readlen + 1);
  assert (end != NULL);

  /* remove newline character from message */
  if (end > readline && end[-1] == '\n')
    end--;
  if (end > readline && end[-1] == '\r')
    end--;

  ret = ceelog_event_new_from_string (readline, end - readline);

  free (readline);
  return ret;
}

static const struct ceelog_source_ops file_ops = {
  file_close,
  file_read_event,
  file_read_event_nonblocking
};

/* Open log file at PATH as an event source.
   Return the event source (for ceelog_source_free ()), or NULL on error. */
struct ceelog_source *
ceelog_source_new_log_file (const char *path, GError ** error)
{
  FILE *f;
  struct ceelog_source *src;

  g_return_val_if_fail (path != NULL, NULL);
  g_return_val_if_fail (error != NULL || *error == NULL, NULL);

  f = fopen (path, "r");
  if (f == NULL)
    {
      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_IO,
		   _("Error opening `%s': %s"), path, strerror (errno));
      return NULL;
    }

  src = ceelog_source_alloc (&file_ops);

  /* FIXME: This should be a full path, to get better error messages */
  src->v.logfile.path = g_strdup (path);
  src->v.logfile.file = f;
  src->v.logfile.prevline = NULL;

  src->v.logfile.inotify_desc = inotify_init ();
  if (src->v.logfile.inotify_desc < 0)
    {
      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_FAILED,
		   _("Inotify initialization failed: %s"), strerror (errno));
      ceelog_source_free (src);
      return NULL;
    }

  src->v.logfile.watch_desc =
    inotify_add_watch (src->v.logfile.inotify_desc, src->v.logfile.path,
		       IN_MODIFY | IN_MOVE | IN_MOVE_SELF);
  if (src->v.logfile.watch_desc < 0)
    {
      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_FAILED,
		   _("Inotify failed to add watch on `%s': %s"),
		   src->v.logfile.path, strerror (errno));
      ceelog_source_free (src);
      return NULL;
    }

  return src;
}
