/* ceelog_source_new_static_file () implementation.

Copyright (C) 2012 Red Hat, Inc. All rights reserved.
This copyrighted material is made available to anyone wishing to use, modify,
copy, or redistribute it subject to the terms and conditions of the GNU General
Public License v.2.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.

Author: Miloslav Trmač <mitr@redhat.com> */
#include <config.h>

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>
#include <glib/gi18n-lib.h>

#include <ceelog/ceelog.h>
#include <ceelog/source.h>

#include "event_private.h"
#include "source_private.h"

/* For now, just use a FILE * and getline().  We may optimize this later if
   necessary. */

/* Close SOURCE and free internal data. */
static void
static_file_close (struct ceelog_source *source)
{
  g_free (source->v.file.path);
  fclose (source->v.file.file);
  free (source->v.file.line);
}

/* Read the next event from SOURCE.

   Return the event (for ceelog_event_free ()) if it is available, or NULL if
   no more events will arrive (without setting ERROR), or NULL on error. */
static struct ceelog_event *
static_file_read_event (struct ceelog_source *source, GError **error)
{
  ssize_t len;
  char *end;

  len = getline (&source->v.file.line, &source->v.file.line_size,
		 source->v.file.file);
  if (len == -1)
    {
      if (ferror (source->v.file.file))
	g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_IO,
		     _("Error reading `%s': %s"), source->v.file.path,
		     strerror (errno));
      /* else EOF, don't set ERROR */
      return NULL;
    }

  /* Do one more scan to make sure there are no embedded NULs left in the
     line. */
  end = memchr (source->v.file.line, '\0', len + 1);
  assert (end != NULL);
  if (end > source->v.file.line && end[-1] == '\n')
    end--;
  if (end > source->v.file.line && end[-1] == '\r')
    end--;
  return ceelog_event_new_from_string (source->v.file.line,
				       end - source->v.file.line);
}

static const struct ceelog_source_ops static_file_ops =
  {
    static_file_close,
    static_file_read_event,
    static_file_read_event /* Non-blocking mode doesn't make a difference. */
  };

/* Open a static text file at PATH as an event source.
   Return the event source (for ceelog_source_free ()), or NULL on error.

   "Static" means that events are not expected to be appended to the file while
   we are reading it, and EOF is final. */
struct ceelog_source *
ceelog_source_new_static_file (const char *path, GError **error)
{
  FILE *f;
  struct ceelog_source *src;

  g_return_val_if_fail (path != NULL, NULL);
  g_return_val_if_fail (error != NULL || *error == NULL, NULL);

  f = fopen (path, "r");
  if (f == NULL)
    {
      g_set_error (error, CEELOG_ERROR, CEELOG_ERROR_IO,
		   _("Error opening `%s': %s"), path, strerror (errno));
      return NULL;
    }

  src = ceelog_source_alloc (&static_file_ops);
  /* FIXME: This should be a full path, to get better error messages */
  src->v.file.path = g_strdup (path);
  src->v.file.file = f;
  src->v.file.line = NULL;
  src->v.file.line_size = 0;
  return src;
}
